#!/usr/bin/perl -sw -Ietc
use strict;

use Entity qw(
  %ENTITY
  set_num
  set_num_diff
  check
  check_eq
  parse_file
);
use Segment();
use SegmentSet();
use Section();
use Disasm();

#
# getHexEnv
#
sub getHexEnv($)
{
  return hex(Entity::get_env(shift()));
}

#
# global variables
#

my $ELF_BASE = getHexEnv('TEVWH_ELF_BASE');
my $ELF_ALIGN = getHexEnv('TEVWH_ELF_ALIGN');
my $UNAME = Entity::get_env('TEVWH_UNAME');
my $ARCH = Entity::get_env('TEVWH_ARCH');
my $OUT = Entity::get_env('TEVWH_OUT');
my $TMP = Entity::get_env('TEVWH_TMP');

#
# credits
#
sub credits($;$)
{
  my $credits = shift;
  my $line = join(', ', sort(@$credits));
  $line =~ s/\s*\n//g;
  $ENTITY{'credits'} = $line;

  return undef;
}
#
# addr_of_main
#
sub addr_of_main($;$)
{
  my $line = shift;
  my $base = shift;

  for $_(@$line)
  {
    if (s/^main_l=//)
    {
      set_num $base . '.addr.main', hex($_);
      return undef;
    }
  }
  return "main_l=ADDR expected." 
}

#
# od_entry_point
#
sub od_entry_point($;$)
{
  my $line = shift;
  my $base = shift;

  return "\$#\$line == $#$line" if ($#$line != 0);
  my @number = split(/ /, $$line[0]);
  return "\$#number == $#number" if ($#number != 1);

  my $number = $number[0];
  $number =~ s/ *0x//;
  my $start = hex($number);

  set_num $base . '.start', $start;
  my $ofs = $start - $ELF_BASE;
  set_num $base . '.ofs', $ofs;

  # this is a superflous paranoia test.
  # $start and $ofs _are_ allowed to differ.
  if ($ARCH eq 'i386')
  {
    ($start == 0x08048080) || return "\$start == $start";
    ($ofs == 0x80) || return "\$ofs == $ofs";
  }
  elsif ($ARCH eq 'sparc')
  {
    ($start == 0x10074) || return "\$start == $start";
    ($ofs == 0x74) || return "\$ofs == $ofs";
  }

  return undef;
}

#
# nm_write
#
sub nm_write($;$)
{
  my $line = shift;
  my $base = shift() . '.nm.';

  for $_(@$line)
  {
    next if (!(m/\s_+write\b/));
    my @number = split;
    return "\$#number == $#number" if ($#number != 2);

    set_num $base . 'value', hex($number[0]);
    $ENTITY{$base . 'name'} = $number[2];
    return undef;
  }
  return 'No symbol matching "libc_write" found.';
}

#
# ls
#
sub ls($$)
{
  my $line = shift;
  my $base = shift;

  return "\$#\$line == $#$line" if ($#$line < 0);
  $line = $$line[0];
  return "Expected -rw or -r-\ngot [$line]" if (!($line =~ m/^-r[w-]/));
  my @number = split(/ +/, $line);
  return "\$#number == $#number" if ($#number < 4);
  set_num $base . '.size', $number[4];
  return undef;
}

#
# disasm
#
sub disasm($$)
{
  my $ref_line = shift;
  my $base = shift;

  my $d = Disasm->new($ARCH, $base);
  return $d->disasm($ref_line);
}

#
# scanner_dist
#
sub scanner_dist($$)
{
  my $line = shift;
  my $base = shift;

  my @assignment = split /[; =]+/, $$line[$#$line];
  return "\$#assignment == $#assignment" if ($#assignment != 9);

  set_num $base . '.files', $assignment[1];
  set_num $base . '.det_page', $assignment[3];
  set_num $base . '.det_align', $assignment[5];
  set_num $base . '.min', hex($assignment[7]);
  set_num $base . '.max', hex($assignment[9]);
  set_num $base . '.available', $assignment[1] - $assignment[3];

  return undef;
}

#
# gdb_stack
#
sub gdb_stack($$)
{
  my $line = shift;
  my $base = shift() . '.gdb.stack';

  my $frame;
  my $value = 0;
  for $_(@$line)
  {
    if (s/^\(gdb\)\s+#// || s/^#//)
    {
      my @word = split;
      $frame = $base . '.frame' . $word[0];
      my $prefix = $frame . '.ret';
      set_num $prefix, hex($word[1]);
      $ENTITY{$prefix} = $word[3];
    }
    elsif (s/^\s*at\s+//)
    {
      my @word = split /:/;

      my $prefix = $frame . '.source';
      set_num $prefix, $word[1];
      $ENTITY{$prefix} = $word[0];
    }
    elsif (s/^\(gdb\) 0x([0-9A-Fa-f]+):\s+/$1 /)
    {
      my @number = split;
      my $addr = hex($number[0]);

      for my $number(@number[1..$#number])
      {
        my $prefix = $base . $value++;
        set_num $prefix, hex($number);
	set_num $prefix . '.addr', $addr;
	$addr += 4;
      }
    }
  }
  return undef;
}

#
# segments
#
sub segments
{
  my $ref_line = shift;
  my $base = shift;
  my $readelf = shift;
  my $filename = shift;

  my $msg = ls($ref_line, $base);
  return 'ls: ' . $msg if defined($msg);

  my $set = new SegmentSet($readelf);
  $msg = $set->read($ref_line);
  return $msg if defined($msg);

  my $entry_point = $set->{ENTRY_POINT};
  set_num $base . '.entry.point', $entry_point;
  set_num $base . '.start.ofs', $entry_point - $ELF_BASE;
  
  my %count_type;
  my $ref_segment = $set->{SEGMENT};
  for my $segment(@$ref_segment)
  {
    my $type_name = $segment->{TypeName};
    my $count = ++$count_type{$type_name};
    my $name = $base . '.' . $type_name . $count;

    my $offset = hex($segment->{Offset});
    my $virtaddr = hex($segment->{VirtAddr});
    my $filesiz = hex($segment->{FileSiz});
    my $memsiz = hex($segment->{MemSiz});

    set_num $name . '.offset', $offset;
    set_num $name . '.virtaddr', $virtaddr;
    set_num $name . '.filesiz', $filesiz;
    set_num $name . '.memsiz', $memsiz;
    set_num $name . '.virt.end', $virtaddr + $memsiz;
    set_num $name . '.file.end', $offset + $filesiz;
    set_num $name . '.align', hex($segment->{Align});
  }
  
  my $ref_load = $set->{LOAD};
  my $load1 = $$ref_load[0];
  my $align;
  if (defined($load1))
  {
    $align = hex($load1->{Align});
    return "\$align = $align != $ELF_ALIGN" if ($align != $ELF_ALIGN);

    my $virtaddr = hex($load1->{VirtAddr});
    return sprintf("\$virtaddr = %#x != %#x", $virtaddr, $ELF_BASE)
    if ($virtaddr != $ELF_BASE);

    # 'evil.magic.ofs.D' is set by parse_od()
    my $ofs = $ENTITY{$base . '.ofs.D'};
    if (defined($ofs))
    {
      my $codesize = hex($load1->{FileSiz}) - $ofs;
      set_num $base . '.codesize', $codesize;
      my $bloat = int($codesize) / int(check($base . '.size.D'));

      my $name = $base . '.bloat';
      die "defined($name) [$ENTITY{$name}] vs. [$bloat]"
      if (defined($ENTITY{$name}));

      $ENTITY{$name} = sprintf("%.3f", $bloat);
      $ENTITY{$name . '.percent'} = sprintf("%.0f", $bloat * 100);
    }
  }

  my $load2 = $$ref_load[1];
  if (defined($load2))
  {
    my $name = $base . '.LOAD1.virt.end.D';
    my $end = $ENTITY{$name} || return "undefined $name";
    set_num $base . '.entry.point.dist.to.end', $end - $entry_point;

    my ( $file_diff, $virt_diff ) = $load1->calc_gap($load2);
    $base .= '.LOAD2.';
    set_num $base . 'file.diff', $file_diff;
    set_num $base . 'virt.diff', $virt_diff;
    my $align_diff = $virt_diff - $align;
    set_num $base . 'diff.align', $align_diff if ($align_diff >= 0);
  }
  return undef;
} 

#
# sections
# 
sub sections($$)
{ 
  my $line = shift;
  my $base = shift;
  my $readelf = shift;
  my $filename = shift;

  my $pattern = '\s*[0-9]+';
  my $func = \&Section::from_objdump;
  if ($readelf)
  {
    $pattern = '^\s*\[\s*[0-9]+\]';
    $func = \&Section::from_readelf;
  }
    
  for(my $i = 0; $i < $#$line; $i++)
  {
    $_ = $$line[$i];
    if (m/^$pattern\s+\.text\s/)
    {
      my $s = Section->new;
      my $msg = $s->$func($_, $$line[++$i]);
      die "sections(\"$filename\"): $1: $msg\n" if defined($msg);

      set_num $base . '.text.addr', hex($s->{ADDR});
      set_num $base . '.text.off', hex($s->{OFFSET});
      set_num $base . '.text.size', hex($s->{SIZE});
      return undef;
    }
  }
  return "Expected section .text";
} 

#
# main
#

$ENTITY{'base.addr'} = sprintf('0x%x', $ELF_BASE);

my @basic = (
  [ 'credits', 'CREDITS' ],
  [ 'scanner_dist', "$OUT/scanner/dist_big", 'dist.big' ],
  [ 'addr_of_main', "$OUT/magic_elf/addr_of_main", 'magic.elf' ],
  [ 'disasm', "$OUT/magic_elf/gdb", 'magic_elf.gdb' ],
  [ 'disasm', "$OUT/evil_magic/static_main.gdb", 'magic_elf_static.gdb' ],

  # set evil.magic.ofs.D, required by segments below
  [ 'od_entry_point', "$OUT/evil_magic/e_entry", 'evil.magic' ],

  [ 'nm_write', "$OUT/evil_magic/nm", 'evil.magic' ],
  [ 'segments', "$OUT/readelf/segments/objdump", 'evil.magic', 0 ],
  [ 'sections', "$OUT/readelf/sections/objdump", 'evil.magic', 0 ],
  [ 'segments', "$OUT/readelf/sh/segments/objdump", 'sh', 0 ],
  [ 'sections', "$OUT/readelf/sh/sections/objdump", 'sh', 0 ],
  [ 'segments', "$OUT/segment_padding/objdump", 'infe1', 0 ],
  [ 'disasm', "$OUT/entry_point/sh.gdb", 'sh.gdb' ],
  [ 'ls', "ls -l $TMP/one_step_closer/i1/infection |", 'infection' ],
);
for my $file(@basic)
{
  my $msg = &parse_file(@$file);
  print $msg . "\n" if (defined($msg));
}

if ($OUT =~ m#out/i386-redhat[0-9\.]+-linux#)
{
  my @advanced = (
    [ 'disasm', "$OUT/entry_point/sh.disasm", 'sh.disasm' ],
    [ 'segments', "$OUT/additional_cs/objdump", 'additional.cs', 0 ],
    [ 'disasm', "$OUT/entry_point/e2.disasm", 'infe2.disasm' ],
    [ 'disasm', "$OUT/entry_point/e3.disasm", 'infe3.disasm' ],
    [ 'gdb_stack', "$OUT/stub_revisited/stack", 'evil.magic' ],
  );
  for my $file(@advanced)
  {
    my $msg = &parse_file(@$file);
    print $msg . "\n" if (defined($msg));
  }
  # check 'evil.magic.LOAD1.filesiz.D', 0x97;
  set_num_diff 'additional.cs.size.delta',
    'additional.cs.size.D', 'sh.size.D';
}

check 'sh.entry.point.D', check('sh.text.addr.D');
check 'evil.magic.entry.point.D', check('evil.magic.start.D');

# check_eq 'sh.gdb.call0.name', '__libc_start_main';
#check 'evil.magic.gdb.stack4.D',
#  check('evil.magic.gdb.stack.frame1.ret.D');

set_num_diff 'infe1.size.delta',
  'infe1.size.D', 'sh.size.D';
set_num_diff 'infe1.LOAD1.filesiz.delta',
  'infe1.LOAD1.filesiz.D', 'sh.LOAD1.filesiz.D';
set_num_diff 'infe1.LOAD2.offset.delta',
  'infe1.LOAD2.offset.D', 'sh.LOAD2.offset.D';
set_num_diff 'infe1.DYNAMIC1.offset.delta',
  'infe1.DYNAMIC1.offset.D', 'sh.DYNAMIC1.offset.D';

open(XML, "> $::xml") || die "$!";
for my $name(sort(keys(%ENTITY)))
{
  printf XML "<!ENTITY calc.%s \"%s\">\n", $name, $ENTITY{$name};
}
close XML;
